from .device_model import DevicesModel, DeviceStatesModel
from .colors_model import ColorsModel
from .multi_command_devices_model import MultiCommandDevicesModel, CommandsModel
from .multi_command_rgb_lamps_model import MultiCommandRgbLampsModel
from .rgb_light_model import RgbLampsModel, RgbLampStatesModel
from .acs_model import AcsModel, AcStatesModel
from .tvs_model import TvsModel, TvStatesModel
from .plugins_model import PluginsModel
from durc import clone



from durc import Model

class PluginsModel(Model):
    schema = {
        **Model.schema,
        'module_name': {
            'type': 'string',
            'required': True,
        },
        #'activated': { 'type': 'bool',  'default': True, 'required': True },
        'version': { 'type': 'string' },
        'description': { 'type': 'string' },
        'config': {
            'type': 'dict',
            'default': {},
        },
    }



import importlib
from . import Controller
from ..models import PluginsModel

class PluginController(Controller):
    model_class = PluginsModel

    @property
    def device_controller_classes(self): return self.module.device_controllers


    @property
    def module_name(self): return self.model.module_name


    def __init__(self):
        self.module = None
        self.ctrl_cls = {}
        self.controllers = {}


    def on_load(self, **data):
        pass


    def init(self):
        self.module = importlib.import_module(self.module_name)
        if 'on_load' in list(self.module.__dict__.keys()):
            self.module.on_load(**self.model.config)
        for ctrl_cls in self.device_controller_classes:
            self.ctrl_cls[ctrl_cls.model_class.collection] = ctrl_cls
        children_model_classes = [(k, v.model_class) for k, v in self.ctrl_cls.items()]
        self.model.on_children_classes(children_model_classes)
        self.model.load_children()
        for col_id, models in self.model.children_models.items():
            if col_id in self.ctrl_cls:
                for m in models:
                    device_controller = self.ctrl_cls[col_id]()
                    device_controller.set_model(m)
                    self.controllers[device_controller.id] = device_controller



from durc import Controller
from .rgb_lamp_controller import RgbLampController
from .device_controller import DeviceController
from .plugin_controller import PluginController
from .tv_controller import TvController
from .ac_controller import AcController
from .multi_command_device_controller import MultiCommandDeviceController
from .multi_command_rgb_lamp_controller import MultiCommandRgbLampController



from .device_model import DevicesModel, DeviceStatesModel
from durc import clone


class TvStatesModel(DeviceStatesModel):
    schema = clone(DeviceStatesModel.schema) | {
        'volume': { 'type': 'integer', 'min': 0, 'max': 100, 'default': 0 },
        'mute': { 'type': 'boolean', 'default': False },
    }
    collection = DeviceStatesModel.collection


class TvsModel(DevicesModel):
    schema = clone(DevicesModel.schema) | {
        'volume_max': { 'type': 'integer', 'default': 100 },
        'volume_min': { 'type': 'integer', 'default': 0 },
    }
    children_model_classes = clone(DevicesModel.children_model_classes)
    children_model_classes['state']['class'] = TvStatesModel



from .rgb_lamp_controller import RgbLampController
from .multi_command_device_controller import MultiCommandDeviceController
from ..models import MultiCommandRgbLampsModel, CommandsModel
from ..models import ColorsModel
from ..utils import rgb_to_hex, hex_to_rgb, color_dist
from math import inf


class MultiCommandRgbLampController(MultiCommandDeviceController, RgbLampController):
    model_class = MultiCommandRgbLampsModel


    def on(self):
        self.send_command('on')
        return True


    def off(self):
        self.send_command('off')
        return True


    def set_brightness_by(self, n):
        self.set_int_state_attr_by('brightness', n)


    def set_brightness(self, target):
        self.set_int_state_attr_to('brightness', target)


    def set_color(self, target_color):
        db = self.model.db
        color_models = []
        for c in self.model.color_commands:
            color_models.append({
                'ctrl': ColorsModel.load(db, c.color_id),
                'cmd_id': c.command_id,
            })
        if len(color_models) == 0: return False
        if type(target_color) == str: target_color = hex_to_rgb(target_color)
        cmd_id = color_models[0]['cmd_id']
        closest_color = [255, 255, 255]
        dist = inf
        for c in color_models:
            ctrl_color = c['ctrl'].rgb
            d = min(color_dist(target_color, ctrl_color), dist)
            if d < dist:
                cmd_id = c['cmd_id']
                dist = d
                closest_color = ctrl_color
        self.send_command(cmd_id)
        self.set_state(color=rgb_to_hex(*closest_color))
        return True



from durc import Model, clone


class ColorsModel(Model):
    schema = {
        **Model.schema,
        'label': { 'type': 'string', 'default': '' },
        'name': { 'type': 'string', 'default': '' },
        'rgb': { 'type': 'list', 'default': [255, 255, 255] },
    }



from abc import ABC, abstractmethod

class Manager(ABC):
    _instance = None


    def __init__(self):
        self._client = None
        self.devices = {}


    @classmethod
    @property
    def instance(cls):
        if not cls._instance:
            cls._instance = cls()
        return cls._instance


    @property
    def client(self): return self._client


    @client.setter
    def client(self, c): self._client = c


    @classmethod
    @abstractmethod
    def init_client(self, **config):
        pass


    @abstractmethod
    def on_device_added(self, device):
        pass


    @classmethod
    def add_device(cls, device):
        i = cls.instance
        i.devices[device.id] = device
        i.on_device_added(device)


    def get_device(self, id):
        if id in self.devices:
            return self.devices[id]
        return None



from math import sqrt


def get_base_classes(cls):
    r = [cls]
    for b in cls.__bases__:
        if b.__name__ != 'object':
            r.extend(get_base_classes(b))
    return r


def get_methods_in(*classes):
    r = []
    for cls in classes:
        methods = [(key, f) for key, f in cls.__dict__.items() if callable(f)]
        r.extend(methods)
    return r


def hex_to_rgb(hex_code):
    h = hex_code.lstrip('#')
    return tuple(int(h[i:i+2], 16) for i in (0, 2, 4))


def rgb_to_hex(*rgb_components):
    r, g, b = rgb_components
    return '#{:02x}{:02x}{:02x}'.format(r, g, b)


def color_dist(a, b):
    dr = b[0] - a[0]
    dg = b[1] - a[1]
    db = b[2] - a[2]
    return sqrt(dr * dr + dg * dg + db * db)
